// Uses natural logarithms everywhere (MK May, 2003)

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "math.h"

#define DEBUG_ON	false	

#define BIC			1
#define ABIC		2
#define SCORETYPE	BIC

#define MAX_MARKERS	128
#define BLOCK	0
#define GAP		1

static int markers;
static FILE *fi, *fo1, *fo2;

static double score[MAX_MARKERS][MAX_MARKERS];

static double prefprob[MAX_MARKERS]; 
static double suffprob[MAX_MARKERS];

int min(int a, int b){if (a < b) return a; return b;}
int max(int a, int b){if (a > b) return a; return b;}

class Solution {
public:
	Solution(){ len = 0; range = new Range[markers+1]; }
	~Solution(){ delete [] range;}
	void append(int bg, int a, int b){
		range[len].bg = bg; range[len].a = a; range[len++].b = b; 
	}
	void print(FILE *f){
		for (int i = 0; i < len; i++){
			if (range[i].bg == BLOCK){
				fprintf(f, " [%d,%d]", range[i].a, range[i].b);
			}
			if (range[i].bg == GAP){
				fprintf(f, " <%d,%d>", range[i].a, range[i].b);
			}
		}
	}
	void println(FILE *f){ print(f); fprintf(f, "\n"); }
	void print2(FILE *f){
		for (int i = 0; i < len; i++){
			if (range[i].bg == BLOCK){
				fprintf(f, " B %d %d", range[i].a, range[i].b);
			}
			if (range[i].bg == GAP){
				fprintf(f, " G %d %d", range[i].a, range[i].b);
			}
		}
	}
	void println2(FILE *f){ print2(f); fprintf(f, "\n"); }
	void sort(){
		// Insertion sort;
		for (int i = len-2; i >= 0; i--){
			Range tmp = range[i];
			int j = i + 1;
			for (j = i + 1; j < len && (tmp.a > range[j].a); j++){
				range[j-1] = range[j];
			} 
			range[j-1] = tmp;
		}
	}
private:
	struct Range {int bg, a, b;}; 
	int len;
	Range *range;
};


void read_data(FILE *f){
	int i, j;
	double s;
	markers = 0;
	if (DEBUG_ON) fprintf(stderr, "optimize > read_data...\n");
	while (fscanf(f, "%d %d %lf", &i, &j, &s) == 3){
		score[i][j] = s;
		if (i > markers) markers = i;
		//if (DEBUG_ON) fprintf(stderr, "%d %d %8.3f \n", i, j, s);
	}
	markers++;
	if (DEBUG_ON) fprintf(stderr, "optimize > %d markers read.\n", markers);
}
void read_data_many(FILE *f){
	int i, j;
	double bics, abics;
	markers = 0;
	if (DEBUG_ON) fprintf(stderr, "optimize > read_data...\n");
	while (fscanf(f, "%d %d %lf %lf", &i, &j, &bics, &abics) == 4){
		switch (SCORETYPE){
		case BIC: score[i][j] = bics; break;
		case ABIC: score[i][j] = abics; break;
		}
		if (i > markers) markers = i;
		//if (DEBUG_ON) fprintf(stderr, "%d %d %8.3f \n", i, j, bics);
	}
	markers++;
	if (DEBUG_ON) fprintf(stderr, "optimize > %d markers read.\n", markers);
}


//
// Optimal solutions are cumulated to a table d[][]. 
// An element d[k][r] tells the optimal decision 
// (si, bg), where 0 <= si <= k, bg = BLOCK or GAP.
//
struct decision {int si; int bg; };
//
static void optimize_with_gapbound(int bound){
	int maxmaxgap = min(markers, bound);

	double b[MAX_MARKERS][MAX_MARKERS];
	double g[MAX_MARKERS][MAX_MARKERS];
	double s[MAX_MARKERS][MAX_MARKERS];
	decision d[MAX_MARKERS][MAX_MARKERS];
	Solution solution[MAX_MARKERS];
	
	// Initialize the score tables b and g. 
	for (int i = 0; i < markers; i++){
		for (int j = i; j < markers; j++){
			b[i][j] = score[i][j];
		    g[i][j] = 0;
		}
	}
	
	// Iterate through k = 0, 1,..., markers-1. 
	for (int k = 0; k < markers; k++){
		for (int u = 0; u <= maxmaxgap; u++){
		// Compute s_u(k) = 
		//  min{ s_u(i-1) + b[i,k], s_{u-(k-i+1)}(i-1) + g[i,k] }.	
			
			double best = 9e99;
			for (int i = 0; i <= k; i++){
				double sb = b[i][k];
				if (i > 0) sb += s[u][i-1];
				if (sb < best){
					best = sb; 
					d[u][k].si = i; d[u][k].bg = BLOCK;
				}
				
				double sg = g[i][k];
				if (u < k-i+1) sg = 9e99;
				else if (i > 0) sg += s[u-(k-i+1)][i-1];
				if (sg < best){
					best = sg;
					d[u][k].si = i; d[u][k].bg = GAP;
				}
			}
			s[u][k] = best; 
			//fprintf(stderr, "u: %d k: %d best: %e \n", u, k, best);
		}
	}
	
	// Iterate backwards to obtain the solutions.
	for (int u = 0; u <= maxmaxgap; u++){
		int k = markers-1, r = u;
		while (k >= 0){
			int si = d[r][k].si, bg = d[r][k].bg;
			solution[u].append(bg, si, k);			
			if (bg == GAP) r -= k - si + 1;
			k = si - 1; 
		}
	}
	
	// Print the solution.
	//fprintf(stderr, "gapmax  opt score   opt solution \n");
	for (int u = 0; u <= maxmaxgap; u++){
		fprintf(fo1, " %4.3f   %8.4e  ",
			u/(double)markers,s[u][markers-1]);
		solution[u].sort();
		solution[u].println2(fo1);
	}
}


// --------
// Bayesian block boundary computations below.
// --------
//
// Returns the joint probability of data D[a, b] 
// and a segment [a, b]. The prior is chosen so 
// that each segmentation of [0, m-1] has the 
// equal prior probability (1/2)^(m-1).
// The likelihood is obtained from the scores.  
//
// The straightforward implementation is numerically unstable. 
// The version belows tackles this problem by using a nontrivial 
// logarithmic transformation.
//
// The score is assumed to be the natural logarithm of 1/P(a,b), 
// i.e., -log P(a,b)
//

#define MARK 9e99
// Substitutes la += log(1+exp(lb-la)) <=> a = a + b.
#define LOGADD(la,lb) if (la>8e99 || lb-la>100) la=lb; else la+=log(1+exp(lb-la));

double log_prob_primitive(int a, int b){
	double log_likelihood = score[a][b] * (-1.0);
	return log_likelihood;
}
void log_boundary_probabilities(){	
	// Compute by dynamic programming.
	for (int k = 0; k < markers; k++){
		prefprob[k] = MARK;
		LOGADD(prefprob[k], log_prob_primitive(0, k));
		for (int r = 0; r < k; r++){
			// Consider the partition 
			// (divisible[0, r], primitive[r+1,k]) of divisible[0, k].
			// Using true probabilities this is:
			// prefprob[k] += prefprob[r] * prob_primitive(r+1, k);
			// A logarithmic trasformation gives
			// x := x + y*z <=> 
			// log x := log x + log(1 + e^{log y + log z - log x})
			//#double w = prefprob[r] + log_prob_primitive(r+1,k) - prefprob[k];
			// An innocent little approximation is used...
			//#if (w > 100) prefprob[k] += w;
			//#else prefprob[k] += log(1 + exp(w));	
			LOGADD(prefprob[k], prefprob[r] + log_prob_primitive(r+1,k));
		}
		//fprintf(stderr, "prefprob[%2d] = %e \n", k, prefprob[k]);
	}
	for (int k = markers-1; k >= 0; k--){
		suffprob[k] = MARK;
		LOGADD(suffprob[k], log_prob_primitive(k, markers-1));
		for (int r = k; r < markers-1; r++){
			// Consider the partition 
			// (primitive[k, r], divisible[r+1,k]) of divisible[k, markers-1].
			// Using true probabilities this is:
			// suffprob[k] += prob_primitive(k, r) * suffprob[r+1];
			// For the following, see above. 
			//#double w = suffprob[r+1] + log_prob_primitive(k,r) - suffprob[k];
			// An innocent little approximation is used...
			//#if (w > 100) suffprob[k] += w;
			//#else suffprob[k] += log(1 + exp(w));	
			LOGADD(suffprob[k], suffprob[r+1] + log_prob_primitive(k,r));
		}
		//fprintf(stderr, "suffprob[%2d] = %e \n", k, suffprob[k]);
	}
	// Summarize the computations and print the results.
	
	double sum = prefprob[0] + suffprob[1] - prefprob[markers-1];
	
	for (int k = 0; k < markers-1; k++){
		double p = exp(prefprob[k] + suffprob[k+1] - prefprob[markers-1]);
		
		if (k > 0){
			double w = prefprob[k] + suffprob[k+1] 
				- prefprob[markers-1] - sum;
			if (w > 10) sum += w;
			else sum += log(1 + exp(w));
		}
		double lodds = log(p) - log(1 - p); 
		if (1 - p < 1e-99) lodds = 100.0;
		if (p < 1e-299){ p = 1e-299; lodds = -999; }
		fprintf(fo2, "%4.2f %e %e \n", k+1.5, 100 * p, lodds);
	}
	
	fprintf(stderr, "Sum of boundary probabilities: %5.3f\n", exp(sum));
	if (DEBUG_ON){
		fprintf(stderr, "optimize > check Bayesian probs: %e = %e ? \n", 
			prefprob[markers-1], suffprob[0]);
	}	
}
// --------
// Bayesian block boundary computations above.
// --------
//
//
void number_of_blocks(){
	
	double logpr[MAX_MARKERS+1][MAX_MARKERS+1];
	
	for (int j = 0; j < markers; j++)
	for (int k = 1; k <= markers; k++) logpr[k-1][j] = 0;
	
	// Position. 
	for (int j = 0; j < markers; j++){
	
		// Number of blocks.
		for (int k = markers; k > 0; k--){
			// If k==1 then logpr[k-1][i] := 0 for all i. 
			// First logpr[k][j] := -oo. 
			// So, first put log x := log y + log z; 
			
			
			if (k == 1){
				logpr[k][j] = log_prob_primitive(0,j);
			}
			else {
			logpr[k][j] = logpr[k-1][k-1] + log_prob_primitive(k,j);
			// Last block.
			for (int i = k; i <= j; i++){
				// Q(k,j) += Q(k-1,i-1)*q(i,j)
				// A logarithmic transformation gives
				// x := x + y*z <=> 
				// log x := log x + log(1 + e^{log y + log z - log x})
				double w = logpr[k-1][i-1] +
					log_prob_primitive(i,j) - logpr[k][j];
				// An innocent little approximation is used...
				
				if (w > 10) logpr[k][j] += w;
				else logpr[k][j] += log(1 + exp(w));
				
			}
			}
		}
	/*
	fprintf(stderr, 
	" At %d-th marker: Number of blocks K: P(K = k) for k = 0 to 10: \n", 
		j+1);
	for (int k = 0; k < 10; k++){
		fprintf(stderr, " %e", logpr[k][j]);
	}
	fprintf(stderr, "\n");	
	*/
	}
	fprintf(stderr, 
	" Number of blocks K: P(K = k | Data) for k = 1 to 9: \n");
	for (int k = 1; k < 10; k++){
		fprintf(stderr, " %1.2e", exp(logpr[k][markers-1] - suffprob[0]));
	}
	fprintf(stderr, "\n");
	//for (int k = 1; k < 10; k++){
	//	fprintf(stderr, " %1.2e", logpr[k][markers-1] - suffprob[0]);
	//}
	//fprintf(stderr, "\n");

}



int main(int argc, char **argv){
	char fin[64], fo1n[64], fo2n[64];
	int i = 0;
	while (argv[2][i] != '\0'){
		fin[i] = fo1n[i] = fo2n[i] = argv[2][i];
		i ++;
	}
	fin[i] = fo1n[i] = fo2n[i] = '.';
	fin[i+1] = 's'; fin[i+2] = 'c'; fin[i+3] = 'o'; fin[i+4] = '\0';
	fo1n[i+1] = 'b'; fo1n[i+2] = 'l'; fo1n[i+3] = 'o'; fo1n[i+4] = '\0';
	fo2n[i+1] = 'm'; fo2n[i+2] = 'b'; fo2n[i+3] = 'p'; fo2n[i+4] = '\0';
	
	fprintf(stderr, "optimize > I/O: %s -> %s, %s\n", fin, fo1n, fo2n); 
	
	fi = fopen(fin, "r");
	read_data_many(fi);
	fclose(fi);
	
	fo1 = fopen(fo1n, "w");
	fo2 = fopen(fo2n, "w");
		
	optimize_with_gapbound(atoi(argv[1]));
	
	log_boundary_probabilities();
	
	number_of_blocks();
	
	fclose(fo1); 
	fclose(fo2);
	
	return 1;
}
